module Export_Functions

  use Constants
  use Basic_IO, only: new_unit
  use Spectral_Element
  use Boundary_Segment
  use Boundary_Curve
  use Boundary_Condition

  implicit none
  private

  type, public :: ExportControlData
    character(len=80) :: casename = 'default'
  end type ExportControlData

  type(ExportControlData), public, save :: ExportData

  public :: GetExportControlData
!  public :: ExportBoundarySplineToGnuplot
  public :: ExportBoundaryVerticesToGnuplot
  public :: ExportBoundaryControlPointsToGnuplot

  public :: ExportGridToSemtex
  public :: CountBCKeys
  public :: CountSurfaces
  public :: CountCurves
  public :: CalculateRadius
  public :: FindLocalEdgeNumber

  character(len=20), parameter :: fmt1 = '(1ES13.5)'
  character(len=20), parameter :: fmt2 = '(2ES20.10)'
  character(len=20), save :: fmtT = ''

contains

!===============================================================================

subroutine GetExportControlData(unit)
  integer, intent(in) :: unit
  character(len=80)   :: casename
  integer	      :: ios
  namelist/exportcase/ casename
  casename = ExportData%casename
  rewind(unit)
  read(unit,nml=exportcase,IOSTAT=ios)
  ExportData%casename = casename
end subroutine GetExportControlData

!===============================================================================

!subroutine ExportBoundarySplineToGnuplot(edge,bseg,curve)
!  type(SpectralElementEdge), intent(in)	:: edge(:)
!  type(BoundarySegment), intent(in)	:: bseg
!  type(BoundaryCurve), intent(in)	:: curve
!  character(len=80)			:: file
!  integer				:: i, j, unit, ios, K
!  real(RNP)				:: t, Kreal, x(2)

! Änderung 16.07.2007
!  K     = 50
!  K     = curve%points !K = Anzahl der zusätzlichen Punkte (außer Knotenpunkten)?
! ---------------------
!  Kreal = K
!  file = exportdata%casename
!  write(file(len_trim(file)+1:),'(3A)') &
!  '_bndry-spline_', bseg%name(1:len_trim(bseg%name)), '.dat'
!
!  write(*,'(2A)')   '  Writing the boundary spline:'
!  write(*,'(2A)') '  ... ', file
!  write(*,'(A)')  repeat('*',45)
!  unit = new_unit()
!  open(unit, IOSTAT=ios, file=file, action='WRITE', status='REPLACE')
!  if (ios /= 0) STOP 'An error occurred while opening the file.'
!  do i = 1, size(bseg%Edge)
!    do j = -K, K
!      t = j / Kreal
!      call BSPoint(curve,edge(bseg%edge(i)%e)%geo%t,x,t)
!      write(unit,fmt2) x
!    end do
!  end do
!  close(unit)
!  write(*,'(A,/,A)') '  done', repeat('*',45)

!end subroutine ExportBoundarySplineToGnuplot

!===============================================================================

subroutine ExportBoundaryVerticesToGnuplot(vertex, edge, bseg)
  type(SpectralElementVertex), intent(in) :: vertex(:)
  type(SpectralElementEdge), intent(in)	  :: edge(:)
  type(BoundarySegment), intent(in)	  :: bseg
  character(len=80)			  :: file
  integer				  :: i, unit, ios

  file = exportdata%casename
!  write(file(len_trim(file)+1:),'(3A)') &
!  '_bndry-vertices_', bseg%name(1:len_trim(bseg%name)), '.dat'
  write(file(len_trim(file)+1:),'(3A)') &
  '_bndry_', bseg%name(1:len_trim(bseg%name)), '.dat'

!  write(*,'(2A)')   '  Writing the boundary vertices:'
  write(*,'(2A)')   '  Writing the boundary geometry file:'
  write(*,'(2A)') '  ... ', file
  write(*,'(A)')  repeat('*',45)
  unit = new_unit()
  open(unit, IOSTAT=ios, file=file, action='WRITE', status='REPLACE')
  if (ios /= 0) STOP 'An error occurred while opening the file.'
  write(unit,fmt2) vertex(edge(bseg%edge(1)%e)%v(1))%x
  do i = 1, size(bseg%Edge)
    write(unit,fmt2) vertex(edge(bseg%Edge(i)%e)%v(2))%x
  end do
  close(unit)
  write(*,'(A,/,A)') '  done', repeat('*',45)

end subroutine ExportBoundaryVerticesToGnuplot

!===============================================================================

subroutine ExportBoundaryControlPointsToGnuplot(bseg,curve)
  type(BoundarySegment), intent(in) :: bseg
  type(BoundaryCurve), intent(in)   :: curve
  character(len=80)		    :: file = ''
  integer			    :: i, unit, ios

  file = exportdata%casename
!  write(file(len_trim(file)+1:),'(3A)') &
!  '_bndry-ctrl-points_', bseg%name(1:len_trim(bseg%name)), '.dat'
  write(file(len_trim(file)+1:),'(3A)') &
  '_bndry_', bseg%name(1:len_trim(bseg%name)), '.dat'

!  write(*,'(2A)') '  Writing the boundary control points:'
  write(*,'(2A)')   '  Writing the boundary geometry file:'
  write(*,'(2A)') '  ... ', file
  write(*,'(A)')  repeat('*',45)
  unit = new_unit()
  open(unit, IOSTAT=ios, file=file, action='WRITE', status='REPLACE')
  if (ios /= 0) STOP 'An error occurred while opening the file.'
  do i = 0, size(curve%xc(:,1)) - 1
    write(unit,fmt2) curve%xc(i,:)
  end do
  close(unit)
  write(*,'(A,/,A)') '  done', repeat('*',45)

end subroutine ExportBoundaryControlPointsToGnuplot

!===============================================================================

subroutine ExportGridToSemtex(elmt,vertex,edge,bseg,curve,bcond)

  type(SpectralElement), intent(in)	  :: elmt(:)
  type(SpectralElementVertex), intent(in) :: vertex(:)
  type(SpectralElementEdge), intent(in)	  :: edge(:)

  type(BoundarySegment), pointer	  :: bseg(:)
  type(BoundaryCurve), pointer		  :: curve(:)
  type(BoundaryCondition), pointer	  :: bcond(:)

  character(len=25)                       :: file, curvetype
  integer                                 :: unit, i, ii, j, k, ios
  integer                                 :: element, p1, p2, n_edges, nsurf, ncurv
  integer                                 :: i2, localedge, localedgepartner
  character, allocatable                  :: keylist(:)
  integer,   allocatable                  :: temp_no(:)
  logical                                 :: flag, boundaryknown

  file = exportdata%casename
  write(file(len_trim(file)+1:),'(A)')


  write(*,'(A)')  '  Writing the gridfile:'
  write(*,'(2A)') '  ... ', file
  write(*,'(A)')  repeat('*',45)

  unit = new_unit()
  open(unit, IOSTAT=ios, file=file, action='WRITE', status='REPLACE')
  if (ios /= 0) STOP 'An error occurred while opening the SemTex file.'

!--Write header to session file-------------------------------------------------
  write(unit,'(A)')  repeat('#',80)
  write(unit,'(2A)') '# session file: ', trim(file)
  write(unit,'(A)')  repeat('#',80)
  write(unit,'(A)') ''
!--Write USER section to session file-------------------------------------------
  write(unit,'(A)') '<USER>'
  write(unit,'(A)') '# Initial conditions go here'
  write(unit,'(A)') '</USER>'
  write(unit,'(A)') ''
!--Write FIELDS section to session file-----------------------------------------
  write(unit,'(A)') '<FIELDS>'
  write(unit,'(A)') '# Specification of field variables'
  write(unit,'(A)') '</FIELDS>'
  write(unit,'(A)') ''
!--Write TOKENS section to session file-----------------------------------------
  write(unit,'(A)') '<TOKENS>'
  write(unit,'(A)') '# Solver parameters go here'
  write(unit,'(A)') '</TOKENS>'
  write(unit,'(A)') ''

!--Determine number of boundary group keys--------------------------------------
  allocate(keylist(1:size(bcond)))
  k = CountBCKeys(bcond, keylist)
  if (k > 0) then
!--Write GROUPS section to session file-----------------------------------------
    write(unit,'(A,I0,A)') '<GROUPS NUMBER=', k,'>'
    do i = 1, k
        write(unit,'(I6,A6,A5,A)')  i, keylist(i), '', 'group_name'
    end do
    write(unit,'(A)') '</GROUPS>'
    write(unit,'(A)') ''
!--Write BCS section to session file--------------------------------------------
    write(unit,'(A,I0,A)') '<BCS NUMBER=', k, '>'
    do i = 1, k
      write(unit,'(I6,A6,A5,A)')  i, keylist(i), '', '4'
      write(unit,'(A)') '                 <D>     u = 0      </D>'
      write(unit,'(A)') '                 <D>     v = 0      </D>'
      write(unit,'(A)') '                 <D>     w = 0      </D>'
      write(unit,'(A)') '                 <H>     p = 0      </H>'
    end do
    write(unit,'(A)') '</BCS>'
    write(unit,'(A)') ''
  end if
  deallocate(keylist)

!--Write NODES section to session file------------------------------------------
  write(unit,'(A,I0,A)') '<NODES NUMBER=',size(vertex),'>'
  do i = 1, size(vertex)
    write(unit,'(I6,A,3ES20.10)') i ,' ', vertex(i)%x, 0.
  end do
  write(unit,'(A)') '</NODES>'
  write(unit,'(A)') ''
!--Write ELEMENTS section to session file---------------------------------------
  write(unit,'(A,I0,A)') '<ELEMENTS NUMBER=',size(elmt),'>'
  do i = 1, size(elmt)
    write(unit,'(I6,A,4I6,A)') i,'   <Q>', elmt(i)%v, '   </Q>' 
  end do
  write(unit,'(A)') '</ELEMENTS>'
  write(unit,'(A)') ''

!--Write SURFACES section to session file---------------------------------------
  allocate(temp_no(1:size(bseg))) 

  write(unit,'(A,I0,A)') '<SURFACES NUMBER=', CountSurfaces(bcond,bseg,temp_no),'>'

  nsurf = 1
  do i = 1, size(edge)
! Check if edge(i) is a boundary edge: edge(i)%l(2) == the second neighbor element must be 0
    if (edge(i)%l(2) == 0 ) then 
! Determine edge number (local numbering 1..4)
      localedge = FindLocalEdgeNumber(elmt(edge(i)%l(1)),edge(i))
! If edge is part of a periodic boundary, the corresponding edge must be found
      if (bcond(edge(i)%b(1))%periodic_link /= '') then
! Boundary is periodic, check if it is recorded in "temp_no"
        boundaryknown = .false.
        do j = 1, size(temp_no)
          if (edge(i)%b(1) == temp_no(j) ) then
! Boundary has been recorded before
            boundaryknown = .true.
            exit
          end if
        end do
        if (boundaryknown .eqv. .false.) then
! New boundary, add it to file
! Looking for corresponding boundary
          p1 = edge(i)%b(1) !<- p1 == Number of first boundary condition/segment
          flag = .false.
          do ii = 1, size(bcond)
            if (bcond(edge(i)%b(1))%periodic_link == bcond(ii)%name) then
              if (bcond(ii)%periodic_link /= '') flag = .true.
              p2 = ii !<- Number of second boundary condition/segment
              exit
            end if
          end do
      if (flag .eqv. .false.) then
         write(*,'(A,25A,A)') 'Error: Link not set correctly for periodic boundary "',&
         & trim(bcond(p1)%name), '"'
         write(*,'(A,25A,A)') '... or linked boundary "', trim(bcond(ii)%name),&
         '" is not set to periodic.'
         stop
      else
! "p2" is the number of the corresponding periodic boundary
! It is assumed that both boundaried have the same number of edges,
! but the corresponding edges are numbered in opposite order
        n_edges =  size(bseg(p1)%edge)
! edge(i)%b(2)
! Corresponding "local" edge number on p2 is (-ii + n_edges + 1)
        i2 = -edge(i)%b(2) + n_edges + 1
! "bseg(p2)%edge(i2)%e" is global edge number on segment p2 for "local" edge number i2
! Determine edge number (local numbering 1..4 in element), "localedgepartner" is the partner edge in local numbering to "localedge"
        localedgepartner = &
        FindLocalEdgeNumber(elmt(edge(bseg(p2)%edge(i2)%e)%l(1)),edge(bseg(p2)%edge(i2)%e))
! Write entry for periodic edge
        write(unit,'(3I6,A,2I6,A)') nsurf, edge(i)%l(1), localedge, '   <P>   ',&
        edge( bseg(p2)%edge(i2)%e )%l(1), localedgepartner, '   </P>'
        nsurf  = nsurf + 1
      end if
    end if
      else
! Entry for nonperiodic edge
! Assign boundary condition key to edge(i)
        write(unit,'(3I6,3A)') nsurf, edge(i)%l(1), localedge, '   <B>        ',&
        bcond( edge(i)%b(1) )%key, '      </B>'
        nsurf = nsurf + 1
      end if
    end if
  end do

  deallocate(temp_no)

  write(unit,'(A)') '</SURFACES>'
  write(unit,'(A)') ''

!--Determine number of edges that lie on curved boundaries----------------------
  ncurv = CountCurves(curve,bseg)
  if (ncurv > 0) then !This section is only written if any curves are found
!--Write CURVES section to session file-----------------------------------------
    write(unit,'(A,I0,A)') '<CURVES NUMBER=', ncurv,'>'

    ncurv = 0 ! Reset curve counter
! Go through all boundary segments == go through all specified curves 
! Index of boundary segment == index of boundary curve ?
    do i = 1, size(bseg)
! Determine what kind of curve (straight, spline, arc, etc.) the given boundary segment has
    select case (curve(i)%geom)
      case('spline') ! Cubic splines
        curvetype = 'SPLINE'
        ! Go through all edges on given boundary segment
        do j = 1, size(bseg(i)%edge)
          ncurv = ncurv + 1
          ! Determine edge number (local numbering 1..4)
          localedge = FindLocalEdgeNumber(elmt(edge(bseg(i)%edge(j)%e)%l(1)),edge(bseg(i)%edge(j)%e))
! If no external file has been specified for the spline, a new file is
! generated from element nodes on the curve. This is done afterwards in 'gambit2semtex.f'
! by calling 'ExportBoundaryVerticesToGnuplot'. The names of this file is
! recorded here alongside the respective CURVES
          if (curve(i)%file == '') then ! If no file has been specified...
              write(unit,'(3I6,10A)') ncurv, edge(bseg(i)%edge(j)%e)%l(1), localedge ,&
              '      <',trim(curvetype),'>   ',trim(exportdata%casename),'_bndry_', &
              trim(bseg(i)%name),'.dat','   </',trim(curvetype),'>'
          else
! If an external file has been specified for the spline, the original file information is written to a new file (2D-format) by calling 'ExportControlPointsToGnuplot' in 'gambit2semtex.f'.
            write(unit,'(3I6,10A)') ncurv, edge(bseg(i)%edge(j)%e)%l(1), localedge ,&
            '      <',trim(curvetype),'>   ',trim(exportdata%casename),'_bndry_', &
            trim(bseg(i)%name),'.dat','   </',trim(curvetype),'>'
          end if
        end do
      case('arc') ! Circular Arc
        curvetype = 'ARC'
! Go through all edges on given boundary segment
        do j = 1, size(bseg(i)%edge)
          ncurv = ncurv + 1
! Determine edge number (local numbering 1..4)
          localedge = FindLocalEdgeNumber(elmt(edge(bseg(i)%edge(j)%e)%l(1)),edge(bseg(i)%edge(j)%e))
! Write radius of curvature (with sign) to session file ------------------------
          write(unit,'(3I6,3A,ES17.10,3A)') ncurv, edge(bseg(i)%edge(j)%e)%l(1), localedge,&
          '      <',trim(curvetype),'>   ', CalculateRadius(curve(i),edge,elmt,bseg(i),vertex,j),&
          '   </',trim(curvetype),'>'
        end do
      case default
    ! Unknown kind of curve or straight line, treat boundary curve as straight line
    end select
  end do
  write(unit,'(A)') '</CURVES>'

  end if
  close(unit)

end subroutine ExportGridToSemtex

!===============================================================================

function FindLocalEdgeNumber(element,edge)

  type(SpectralElement), intent(in)         :: element
  type(SpectralElementEdge), intent(in)	    :: edge

  integer                                   :: i
  integer                                   :: FindLocalEdgeNumber

  i = 1
  do
    if (element%v(i) == edge%v(1)) then
        exit
     else
        i = i + 1
     end if
  end do

  FindlocalEdgeNumber = i

end function FindLocalEdgeNumber


!===============================================================================

function CountBCKeys(bcond,keylist)
! Counts the number of different boundary condition keys and stores them in keylist

  type(BoundaryCondition), pointer    :: bcond(:)

  character, intent(out)  :: keylist(:)
  integer                 :: CountBCKeys
  integer                 :: i, j, nkey
  logical                 :: keyfound

  nkey = 0

  do i = 1, size(bcond)
! Periodic boundaries have no group, so they are not counted
    if (bcond(i)%periodic_link /= '') then
      cycle ! Go to next bond(i)
    else
      if (nkey == 0) then
        keylist(1) = bcond(i)%key
        nkey = 1
        keyfound = .true.
      else
        do j = 1, nkey
! Go through all previosly recorded keys
          if (bcond(i)%key == keylist(j)) then
! "keyfound" is true if key has previously been found
            keyfound = .true.
            exit
          else
! key han not yet been found
            keyfound = .false.
          end if
        end do
      end if
! If key has not yet been found, add it to keylist and increase counter nkey
      if (keyfound .eqv. .false.) then
        nkey = nkey + 1
        keylist(nkey) =  bcond(i)%key
      end if
    end if
  end do

  CountBCKeys = nkey

end function CountBCKeys

!===============================================================================

function CountSurfaces(bcond,bseg,temp_no)

  type(BoundaryCondition), pointer    :: bcond(:)
  type(BoundarySegment), pointer      :: bseg(:)

  integer, intent(out)    :: temp_no(:)
  integer                 :: CountSurfaces
  integer                 :: i, j, k, nsurf
  logical                 :: flag

  temp_no = 0 ! Array that
  k = 0
  nsurf = 0

  do i = 1, size(bcond)
! Check if boundary is periodic or not
    if (bcond(i)%periodic_link /= '') then
! Boundary is periodic, check if it is recorded in "temp_no"
      flag = .false.
      do j = 1, size(temp_no)
        if (i == temp_no(j) ) then
! Boundary has been recorded before
          flag = .true.
          exit
        end if
      end do
! If boundary has not been recorded before ...
      if (flag .eqv. .false.) then
! Look for the complementary boundary of "i"
        do j = 1, size(bcond)
          if (bcond(i)%periodic_link == bcond(j)%name) then
! Corresponding boundary "j" found: record it in temp_no
            k = k + 1
            temp_no(k) = j
            exit
          end if
        end do
! Add edges to total
        nsurf = nsurf + size(bseg(i)%edge)
      end if
    else
! Boundary is not periodic, just add all the edges to total
      nsurf = nsurf + size(bseg(i)%edge)
    end if
  end do

  CountSurfaces = nsurf

end function CountSurfaces

!===============================================================================

function CountCurves(curve,bseg)

  type(BoundaryCurve), pointer    :: curve(:)
  type(BoundarySegment), pointer  :: bseg(:)

  integer                 :: CountCurves
  integer                 :: i, ncurv

  ncurv = 0
  do i = 1, size(curve)
    if ( curve(i)%geom /= 'straight') then
    ncurv = ncurv + size(bseg(i)%edge)
    end if
  end do

  CountCurves = ncurv

end function CountCurves

!===============================================================================

function CalculateRadius(curve,edge,elmt,segment,vertex,i)

  type(BoundaryCurve), intent(in)         :: curve
  type(SpectralElementEdge), intent(in)   :: edge(:)
  type(SpectralElement), intent(in)       :: elmt(:)
  type(BoundarySegment), intent(in)       :: segment
  type(SpectralElementVertex), intent(in) :: vertex(:)

  real(RNP)            :: CalculateRadius
  integer              :: i, j, element
  real(RNP)            :: radius, xc, yc, distance

! Radius for edge is average of the radii for each of the two vertices belonging to that edge
! This assumes that both radii have approximately the same value
   radius = 0.5_RNP * (curve%r0(i-1) + curve%r0(i))
! Determine centroid of element attached to edge
   element = edge(segment%edge(i)%e)%l(1)
   xc = 0.0_RNP
   yc = 0.0_RNP
   do j = 1, 4
      xc = xc + vertex(elmt(element)%v(j))%x(1)
      yc = yc + vertex(elmt(element)%v(j))%x(2)
   end do
   xc = xc / 4.0_RNP
   yc = yc / 4.0_RNP
! Determine distance of element centroid from center of circle
   distance = sqrt((xc-curve%x0)**2 + (yc-curve%y0)**2)
! Set sign of radius, positive if convex, negative if concave
   if (distance > radius) radius = -radius

  CalculateRadius = radius

end function CalculateRadius

!===============================================================================

end module Export_Functions
