program Converter

  use Constants
  use Spectral_Element
  use Boundary_Segment
  use Boundary_Curve
  use Boundary_Condition
  use Data_Structure
  use Export_Functions
  use Import_Functions

  implicit none

  type(SpectralElementVertex), pointer	:: vertex(:)
  type(SpectralElementEdge), pointer	:: edge(:)
  type(SpectralElement), pointer	:: element(:)
  type(BoundarySegment), pointer	:: bseg(:)
  type(BoundaryCurve), pointer	        :: curve(:)
  type(BoundaryCondition), pointer      :: bcond(:)

  character(len=80)                     :: casefile
  integer                               :: i

!===============================================================================
! Import
!===============================================================================

  write(*,'(A)') 'specify input file (full name, including extension)'
  read(*,'(A)') casefile

  call GetControlData(casefile) ! controldata from inputfile
  call ImportGrid(vertex,edge,element,bseg) ! mesh import
  do i = 1, size(edge) ! initialisation
    allocate(edge(i)%geo)
  end do
  call ConstructBoundarySegments(vertex,edge,bseg,curve,bcond) ! boundary cnstr.
  call CheckDataStructure(vertex,edge,element,bseg) ! final check

!===============================================================================
! Export
!===============================================================================

!  do i = 1, size(bseg)
!    if ((curve(i)%geom == 'spline') .and. ( curve(i)%file == '')) then
!      if (curve(i)%points == 0) then
!        call ExportBoundaryVerticesToGnuplot(vertex, edge, bseg(i))
!      else
!        call ExportBoundarySplineToGnuplot(edge, bseg(i),curve(i))
!      end if
!    else
!      call ExportBoundaryControlPointsToGnuplot(bseg(i), curve(i))
!    end if
!  end do

!  do i = 1, size(bseg)
!    if (curve(i)%geom == 'spline') then
!      if ( curve(i)%file == '') then
!        if (curve(i)%points == 0) then
!          call ExportBoundaryVerticesToGnuplot(vertex, edge, bseg(i))
!        else
!          call ExportBoundarySplineToGnuplot(edge, bseg(i),curve(i))
!        end if
!      else
!        call ExportBoundaryControlPointsToGnuplot(bseg(i), curve(i))
!      end if
!    end if
!  end do

  do i = 1, size(bseg)
    if (curve(i)%geom == 'spline') then ! If curve(i) is marked as spline...
      if ( curve(i)%file == '') then ! If curve (i) has no geometry source file...
        call ExportBoundaryVerticesToGnuplot(vertex, edge, bseg(i)) ! Create a new geometry
! file using the boundary vertices instead
      else ! If the curve dos have an associated source file...
        call ExportBoundaryControlPointsToGnuplot(bseg(i), curve(i)) ! Take the original data
! and write it into a new geometry file which has only two columns (2D-data)
      end if
    end if
  end do

  call ExportGridToSemtex(element,vertex,edge,bseg,curve,bcond) ! Write out the semtex
! session file

!===============================================================================

end program Converter
