module Spectral_Element

  use Constants
  use Boundary_Curve

  implicit none
  private

  type, public :: SpectralElement
     integer ::  v(4) = 0  ! vertex IDs
     integer ::  e(4) = 0  ! edge IDs
  end type SpectralElement

  type, public :: SpectralElementVertex
     real(RNP) ::  x(2) ! vertex coordinates
  end type SpectralElementVertex

  type, public :: SpectralElementEdgeData
!     logical   :: curved = .false. ! whether edge is curved or straight
     real(RNP) :: t(2)   = 0       ! curve parameter
  end type SpectralElementEdgeData

  type(SpectralElementEdgeData), public, save :: defaultEdgeData

  type, public :: SpectralElementEdge
     integer :: b(2) = 0  ! boundary ID (0 if none)
     integer :: v(2) = 0  ! vertex IDs
     integer :: l(2) = 0  ! element IDs
     type(SpectralElementEdgeData), pointer :: geo => null()
  end type SpectralElementEdge


  public :: FindUndefinedEdge
  public :: ElementOrientation
  public :: FlipElementOrientation

  public :: SharedNode

contains

!===============================================================================

integer function FindUndefinedEdge(element) result(N)
  type(SpectralElement), intent(in) ::  element
  integer ::  i
  N = 0
  do i = 1, 4
    if (element%e(i) == 0) then
       N = i
       exit
    end if
  end do
end function FindUndefinedEdge

!===============================================================================

logical function ElementOrientation(vertex) result(t1)
  type(SpectralElementVertex), intent(in) ::  vertex(4)
  real(RNP) ::  a(2), b(2)
  logical ::  t2

! .TRUE.  :: anticlockwise orientation
! .FALSE. :: clockwise orientation

  t1 = .TRUE.
  t2 = .TRUE.

  a = vertex(2)%x - vertex(1)%x
  b = vertex(3)%x - vertex(1)%x
  if (a(1)*b(2) - a(2)*b(1) < 0) t1 = .FALSE.

  a = vertex(3)%x - vertex(1)%x
  b = vertex(4)%x - vertex(1)%x
  if (a(1)*b(2) - a(2)*b(1) < 0) t2 = .FALSE.

  if (t1.NEQV.t2) STOP 'Error: Different face orientations.'

end function ElementOrientation

!===============================================================================

subroutine FlipElementOrientation(element)
  type(SpectralElement), intent(inout) ::  element
  element%v = element%v((/ 2,1,4,3 /))
  element%e = element%e((/ 1,4,3,2 /))
end subroutine FlipElementOrientation

!===============================================================================

integer function SharedNode(edge1, edge2) result(v)
  type(SpectralElementEdge), intent(in) ::  edge1, edge2

  v = 0
  if (edge1%v(1) == edge2%v(1)) v = edge1%v(1)
  if (edge1%v(1) == edge2%v(2)) v = edge1%v(1)
  if (edge1%v(2) == edge2%v(1)) v = edge1%v(2)
  if (edge1%v(2) == edge2%v(2)) v = edge1%v(2)

end function SharedNode

!===============================================================================

end module Spectral_Element
